import numpy as np
from scipy.special import softmax, expit
import sys
import time
from tqdm import trange
from multiprocessing import Pool
from functools import partial

class Evaluator:
    def __init__(self, P, Q, X_te, y_te, epsilon, bsize, tem, pos=1, noise_scale=0.0, tK=100, alpha=1.0):
        self.P = P
        self.Q = Q
        self.m = P.shape[0]
        self.n = Q.shape[0]
        self.epsilon = epsilon
        self.bsize = bsize
        self.tem = tem
        self.pos = int(pos)
        self.noise_scale = noise_scale
        self.tK = int(tK)
        self.alpha = alpha
        assert self.tem >= 0 and self.noise_scale >= 0 and self.pos >= 1 and self.tK >= self.pos and self.alpha >= 0 and self.alpha <= 1
        self.X_te, self.y_te = X_te, y_te
        if X_te is not None:
            assert self.X_te.shape[0] == self.P.shape[0]

    def take_action(self, policy):
        actions, pps, topK = [], [], []
        for i in trange(0, self.m, self.bsize, desc='%s'%policy):
            sample_num = self.P[i:i+self.bsize].shape[0]
            if policy == 'is':
                logits = model_predict(self.P[i:i+self.bsize, :], self.Q)
                #logits = np.log(expit(logits))
                noises = np.random.gumbel(scale=self.noise_scale, size=logits.shape)
                logits = logits + noises

                #logit_scale = logits.std(axis=-1)
                #noises = np.random.normal(scale=logit_scale, size=logits.shape[::-1]).T
                #logits = self.alpha * logits + (1 - self.alpha) * noises

                logits *= self.tem
                a, p = importance_sampling(logits, self.pos)
                topK.append(get_topK(logits, self.tK))
            elif policy == 'uniform':
                a, p = uniform(sample_num, self.n, self.pos)
                topK.append(uniform(sample_num, self.n, self.tK)[0])
            else:
                raise ValueError('No such polciy:', self.policy)
            actions.append(a.reshape(sample_num, -1)), pps.append(p.reshape(sample_num, -1))

        return np.vstack(actions), np.vstack(pps), np.vstack(topK)

    def get_emp_reward(self, actions):
        rewards = np.array([np.isin(actions[k, :], self.y_te[k]) for k in range(len(self.y_te))])
        return rewards

    def get_exp_reward(self, writer=None):
        rewards = []
        for i in range(0, self.m, self.bsize):
            logits = self.tem*model_predict(self.P[i:i+self.bsize, :], self.Q)
            probs = softmax(logits, axis=1)
            if writer is not None:
                for r in range(0, probs.shape[0], 100):
                    prob = probs[r]
                    #prob = sorted(probs[r])
                    #counts, limits = np.histogram(prob)
                    counts = (np.array(prob)*100000).astype(np.int)
                    _sum = np.sum(counts*np.arange(len(prob)))
                    sum_sq = np.sum(counts*(np.arange(len(prob))**2))
                    writer.add_histogram_raw(
                            tag='hist_dummy_data', \
                            min=0, \
                            max=len(prob), \
                            num=counts.sum(), \
                            sum=_sum, \
                            sum_squares=sum_sq, \
                            bucket_limits=np.arange(1, len(prob)+1).tolist(), \
                            bucket_counts=counts.tolist(), \
                            global_step=(i*self.bsize + r))
            rewards.extend([1.0*probs[j, :][np.array(y).astype(np.int)].sum() for j, y in enumerate(self.y_te[i:i+self.bsize])])
        return rewards

# return topK idx or 0/1 mask of topK values for each row of m
def get_topK(m, topK, return_mask=False):
    topK_idx = np.argsort(-1*m, axis=1)[:, :topK]
    if return_mask:
        mask = np.zeros_like(m, dtype=np.int)
        mask[np.arange(m.shape[0])[:, None], topK_idx] = 1
        return mask
    else:
        return topK_idx

def model_predict(ps, Q):
    return np.matmul(ps, Q.T)
    #return np.ones((X.shape[0], action_num))

def vectorized_random_choice(probs):
    s = probs.cumsum(axis=1)
    s[:, -1] = 1.0+1e-8
    r = np.random.rand(probs.shape[0],1)
    idx = (s < r).sum(axis=1)
    return idx

def uniform(sample_num, action_num, pos):
    s = np.random.rand(sample_num, action_num)
    idx = np.argsort(s, axis=-1)[:, :pos]
    return idx, np.tile(1./np.arange(action_num, action_num-pos, -1), (sample_num, 1))

def e_greedy(logits, e):
    n = logits.shape[1]
    greedy_prob = 1.0 - (e*(n-1.0)/n)
    uniform_prob = e/n
    idx = np.argmax(logits, axis=1)
    rnd_flag = np.random.rand(logits.shape[0]) < e
    idx[rnd_flag] = np.random.choice(n, rnd_flag.sum())
    propensity = np.ones(logits.shape[0], dtype=np.float)*greedy_prob
    propensity[rnd_flag] = uniform_prob
    return idx, propensity

def pl_ranking(action_num, pos, weights):
    s = weights.sum()
    idx = np.zeros(pos, dtype=np.int) 
    prob = np.zeros(pos) 
    for i in range(pos):
        p = weights/weights.sum()
        try:
            idx[i] = np.random.choice(action_num, replace=False, p=p)
        except:
            print(p, p.sum())
        prob[i] = p[idx[i]]
        weights[idx[i]] = 0.0
    
    return idx, prob

def importance_sampling(logits, pos):
    if pos == 1:
        probs = softmax(logits, axis=1)
        idx = vectorized_random_choice(probs)
        return idx, probs[np.arange(probs.shape[0]), idx]
    else:
        weights = np.exp(logits)
        func = partial(pl_ranking, weights.shape[1], pos)
        with Pool(16) as p:
            res = p.map(func, list(weights))
            idx, prob = zip(*res)
        return np.array(idx), np.array(prob)

def read_action_info(fpath):
    action2feat = []
    with open(fpath, 'r') as f:
        for line in f:
            line = line.strip()
            action2feat.append(line)
    return action2feat

def dump_svm(f, X, y_idx, y_propensity, y_value):
    if not hasattr(f, "write"):
        f = open(f, "w")
    X_is_sp = int(hasattr(X, "tocsr"))
    #y_is_sp = int(hasattr(y, "tocsr"))

    value_pattern = "%d:%.6g"
    label_pattern = "%d:%d:%.16g"
    line_pattern = "%s %s\n"
    
    for i, d in enumerate(zip(y_idx, y_value, y_propensity)):
        if X_is_sp:
            span = slice(X.indptr[i], X.indptr[i + 1])
            row = zip(X.indices[span], X.data[span])
        else:
            nz = X[i] != 0
            row = zip(np.where(nz)[0], X[i, nz])

        s = " ".join(value_pattern % (j, x) for j, x in row)
        labels_str = label_pattern % d 
        feat = (labels_str, s)
        f.write(line_pattern % feat)

if __name__ == "__main__":
    np.random.seed(0)
    M = 10000
    N = 1000
    logits = (2.*np.random.rand(M, N) - 1)*5
    print(logits)
    start_time = time.time()
    print(uniform(M, N, 10))
    print(time.time() - start_time)
    start_time = time.time()
    print(importance_sampling(logits, 10))
    print(time.time() - start_time)

