#include <stdio.h>
#include <stdint.h>
#include <math.h>

#include <omp.h>

dtype inner(const dtype *p, const dtype *q, const int k)
{
    int d;
    dtype product=0;
    for (d = 0; d < k; d++)
        product += p[d]*q[d];
    return product;
}

void axpy(dtype *p, const dtype *q, const int k, const dtype a)
{
    int d;
    for (d = 0; d < k; d++)
        p[d] += a*q[d];
}

void cz(const dtype *p, const dtype *q, const dtype *w, const dtype *h, const int k,
        const int32_t *idx, const int32_t *nz_j,
        const size_t m, const size_t start, const dtype *R, const dtype *S,
        const dtype *y_weight, const double omega,
        const dtype *y_sigma, dtype *ZQ) {

#pragma omp parallel for schedule(dynamic)
    for (size_t i = 0; i < m; i++) {
        const dtype *pi = p+(i+start)*k, *wi = w+(i+start)*k;
        dtype *zq_i = ZQ + i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];
            const dtype Ri = R[start+i];
            const dtype Sj = S[j];

            dtype zij = inner(pi, h+j*k, k) + inner(wi, q+j*k, k);
            zij *= y_sigma[l]*(1-y_sigma[l])*y_weight[l]-(dtype)omega*Ri*Sj;
            axpy(zq_i, q+j*k, k, zij);
        }
    }
}

double cpred_val_sogram(const dtype *p, const dtype *q, const int k,
        const dtype *pt, const dtype *qt, const int k1, const size_t nnz,
        const int32_t *is, const int32_t *js, const double omega,
        const dtype *y_label, dtype *bp, dtype *bq) {

    double p_loss = 0;
#pragma omp parallel for schedule(dynamic) reduction(+:p_loss)
    for (size_t l = 0; l < nnz; l++) {
        const dtype *pl = p+l*k;
        const dtype *ql = q+l*k;

        const dtype *ptl = pt+is[l]*k1;
        const dtype *qtl = qt+js[l]*k1;

        const dtype y_hat = inner(pl, ql, k);
        const dtype y_imp = inner(ptl, qtl, k1);

        dtype diff = y_hat - y_imp;

        double residual = omega/2*diff*diff;

        dtype yz = y_label[l]*y_hat;
        double expyy, y_sigma;
        if (yz > 0){
            expyy = exp(-yz);
            y_sigma = expyy/(1+expyy);
            p_loss += log(1+expyy)-residual;
        }
        else {
            expyy = exp(yz);
            y_sigma= 1/(1+expyy);
            p_loss += -yz+log(1+expyy)-residual;
        }
        dtype bl = -y_label[l]*y_sigma - (dtype)omega*diff;

        for (int d = 0; d < k; d++) {
            bp[l*k+d]  = bl*pl[d];
            bq[l*k+d]  = bl*ql[d];
        }
    }
    return p_loss;
}

double cpred_val(const dtype *p, const dtype *q, const int k,
        const int32_t *idx, const int32_t *nz_j,
        const size_t m, const dtype *R, const dtype *S,
        const dtype *y_weight, const double omega, const double alpha,
        const dtype *y_label, const dtype *y_imp, dtype *y_hat, dtype *y_sigma) {

    double p_loss = 0;
#pragma omp parallel for schedule(dynamic) reduction(+:p_loss)
    for (size_t i = 0; i < m; i++) {
        const dtype *pi = p+i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];

            y_hat[l] = inner(pi, q+j*k, k);
            dtype diff = y_hat[l] - y_imp[l];

            const dtype Ri = R[i];
            const dtype Sj = S[j];
            double residual = omega/2*Ri*Sj*diff*diff;

            dtype yz = y_label[l]*y_hat[l];
            double expyy;
            if (yz > 0){
                expyy = exp(-yz);
                y_sigma[l] = expyy/(1+expyy) + 1e-9;
                p_loss += y_weight[l]*log(1+expyy)-residual;
            }
            else {
                expyy = exp(yz);
                y_sigma[l] = 1/(1+expyy) + 1e-9;
                p_loss += y_weight[l]*(-yz+log(1+expyy))-residual;
            }
        }
    }
    return p_loss;
}

void cBQ(const dtype *q, const int k,
        const int32_t *idx, const int32_t *nz_j,
        const size_t m, const size_t start,
        const dtype *R, const dtype *S, const dtype *y_weight,
        const double omega, const dtype *y_label, const dtype *y_imp, const dtype *y_hat,
        const dtype *y_sigma, dtype *BQ) {

#pragma omp parallel for schedule(dynamic)
    for (size_t i = 0; i < m; i++) {
        dtype *bq_i = BQ + i*k;
        for (size_t l = idx[i]; l < idx[i+1]; l++) {
            const size_t j = nz_j[l];

            const dtype Ri = R[start+i];
            const dtype Sj = S[j];

            dtype diff = y_hat[l] - y_imp[l];
            dtype bij = -y_label[l]*y_sigma[l]*y_weight[l] - omega*Ri*Sj*diff;
            axpy(bq_i, q+j*k, k, bij);
        }
    }
}
