import os
import math
import time

import numpy as np
from scipy.sparse import csr_matrix

import tensorflow.compat.v1 as tf

from utilities import predict, save_PQ, vectorize, inverse_vectorize, calc_norm, zero_tensor, get_cfuncs

import pdb

class General(object):
    def __init__(self, config, sess, network):
        """
        initialize operations and vairables that will be used in newton
        args:
            sess: tensorflow session
            outputs: output of the neural network (pre-softmax layer)
            loss: function to calculate loss
       """
        super(General, self).__init__()
        self.network = network
        self.u, self.v, self.p, self.q = network
        self.sess = sess

        self.f = 0

        self.config = config

        self.param = tf.trainable_variables()

        self.dtype = self.config.dtype
        self.dtype_np = self.config.dtype_np

        model_weight = vectorize(self.param)

        self.reg = tf.Variable(0., dtype=self.dtype, trainable=False)

        self.label_idx = tf.placeholder(tf.int64, shape=[self.config.bsize_i, 2])
        self.weights = tf.placeholder(self.dtype, shape=[self.config.bsize_i])
        self.tpk_idx = tf.placeholder(tf.int64, shape=[self.config.bsize_i, self.config.tpk_num])
        self.tpk_mask = tf.placeholder(self.dtype, shape=[self.config.bsize_i, self.config.tpk_num])
        if self.config.loss_type == 4:
            self.pl_mask = tf.sparse_placeholder(self.dtype, shape=[self.config.bsize_i, self.config.tpk_num])
        else:
            self.pl_mask = tf.sparse_placeholder(self.dtype, shape=[self.config.bsize_i, self.config.n])

        self.adagrad_init_accum_value = 0.1
        self.adagrad_eps = 1e-7

        self.tic = 0
        self.total_running_time = 0
        self.nr_print = 0

        self.amp = 1.0
        self.amp *= self.config.m / self.config.bsize_i


        zeros = tf.zeros(model_weight.get_shape(), dtype=self.dtype)

        # initial variable used in gradient
        self.g = tf.Variable(zeros, dtype=self.dtype, trainable=False, name="g")

        self.ada = tf.Variable(zeros+self.adagrad_init_accum_value, dtype=self.dtype, trainable=False, name="ada")
        self.s = tf.Variable(zeros, dtype=self.dtype, trainable=False, name="s")

        self._lambda = tf.placeholder(dtype=self.dtype, shape=[])

        self.update_model = self._update_model()
        self.gnorm = calc_norm(tf.concat([self.g], axis=0))

        self.cal_reg = self._cal_reg()

        self.zero_g = zero_tensor(self.g)
        self.loss, self.batch_grad = self._cal_loss_grad()
        self.reg_grad = self._add_reg_grad()


        self.update_adagrad = tf.assign(self.ada, self.ada + self.g*self.g)
        self.gradient_direction = tf.assign(self.s, -self.g*tf.math.rsqrt(self.ada+self.adagrad_eps))


    def profile(self, *args, **kwargs):
        if self.config.profile:
            print(*args, **kwargs)

    def _cal_reg(self):
        model_weight = vectorize(self.param)
        reg = (calc_norm(model_weight))**2
        return tf.assign(self.reg, reg*self.config._lambda/2)

    def _mask_past_actions(self, logits, num=100000):
        pl_mask = tf.sparse.to_dense(self.pl_mask)
        return logits + (-(((pl_mask * num) + num) - num))

    def _cal_loss_grad(self):

        if self.config.loss_type in [0, 1, 5, 6]:
            logits = tf.matmul(self.p, tf.transpose(self.q))
            logits = self._mask_past_actions(logits)
            probas = tf.nn.softmax(logits)
            if self.config.loss_type == 0:
                weights = tf.gather_nd(tf.stop_gradient(probas), self.label_idx)*self.weights + self.config.zeta
            elif self.config.loss_type == 5:
                probas_logged = tf.gather_nd(tf.stop_gradient(probas), self.label_idx)
                weights = self.config.slot_num*tf.pow(1.0-probas_logged, self.config.slot_num-1.0)*probas_logged*self.weights
            elif self.config.loss_type == 6:
                weights = tf.gather_nd(tf.stop_gradient(probas), self.label_idx)*self.weights
            else:
                weights = self.weights + self.config.zeta
            loss = -weights * tf.math.log(tf.gather_nd(probas, self.label_idx))
        elif self.config.loss_type == 2:
            logits = tf.matmul(self.p, tf.transpose(self.q))
            logits = self._mask_past_actions(logits)
            max, abs, log, exp = tf.math.maximum, tf.math.abs, tf.math.log, tf.math.exp
            neg_loss = max(logits, 0) + log(1 + exp(-abs(logits)))
            neg_loss = tf.reduce_sum(neg_loss, axis=1)
            pos_logits = tf.gather_nd(logits, self.label_idx)
            pos_loss = (1-self.config.omega)*(max(pos_logits, 0) + log(1+exp(-abs(pos_logits)))) - pos_logits
            loss = (self.weights + self.config.zeta)*(pos_loss + self.config.omega * neg_loss)
        elif self.config.loss_type == 3:
            logits = tf.matmul(self.p, tf.transpose(self.q))
            logits = self._mask_past_actions(logits, self.config.r)
            max, abs, log, exp = tf.math.maximum, tf.math.abs, tf.math.log, tf.math.exp
            neg_loss = 0.5*(logits-self.config.r)*(logits-self.config.r)
            neg_loss = tf.reduce_sum(neg_loss, axis=1)
            pos_logits = tf.gather_nd(logits, self.label_idx)
            pos_loss = max(pos_logits, 0) + log(1+exp(-abs(pos_logits))) - pos_logits - self.config.omega*0.5*(pos_logits-self.config.r)*(pos_logits-self.config.r)
            loss = (self.weights + self.config.zeta)*(pos_loss + self.config.omega * neg_loss)
        elif self.config.loss_type == 4:
            qs = tf.reshape(tf.gather(self.q, tf.reshape(self.tpk_idx, [-1]), axis=0), [-1, self.tpk_idx.shape[1], self.q.shape[1]])
            logits = tf.reduce_sum(tf.multiply(tf.expand_dims(self.p, axis=1), qs), axis=-1)
            logits = self._mask_past_actions(logits)
            probas = tf.nn.softmax(logits)
            probas = tf.reduce_sum(probas*self.tpk_mask, axis=1)
            ignore_mask = tf.greater(probas, 0)
            probas = tf.boolean_mask(probas, ignore_mask)
            weights = tf.stop_gradient(probas)*tf.boolean_mask(self.weights, ignore_mask)
            loss = -weights * tf.math.log(probas)

        loss = tf.math.reduce_sum(loss)*self.amp

        grad = tf.gradients(ys=loss, xs=self.param)
        grad = vectorize(grad)
        update_grad = tf.assign(self.g, self.g+grad)

        return loss, update_grad


    def _add_reg_grad(self):
        model_weight = vectorize(self.param)
        reg_grad = model_weight*self.config._lambda
        return tf.assign(self.g, self.g+reg_grad)

    def print_info(self, k, i, num_segment, trainL, trainR, validL, validR, log_file):
        toc = time.time()
        epoch = k + i/num_segment
        self.total_running_time += toc-self.tic
        f = self.f
        output_str = 'iter: {:.2f} f: {:.3f} time: {:.3f}'.format(epoch, f, self.total_running_time)
        print(output_str)
        print(output_str, file=log_file)
        if validL.feat is not None and epoch == k:
            precision, ndcg, precision_tpk, ndcg_tpk = predict(self.sess, self.network, validL, validR, self.config, self.dtype_np)
            precision = "p@k"+ " " + " ".join([str(round(i*100,3)) for i in precision])
            ndcg = "ndcg@k" + " " + " ".join([str(round(i*100,3)) for i in ndcg])
            precision_tpk = "p@k_tpk"+ " " + " ".join([str(round(i*100,3)) for i in precision_tpk])
            ndcg_tpk = "ndcg@k_tpk" + " " + " ".join([str(round(i*100,3)) for i in ndcg_tpk])
            print("iter: %d "%k + precision + "\t" + ndcg + "\t" + precision_tpk + "\t" + ndcg_tpk)
            if not self.config.screen_log_only:
                print("iter: %d "%k + precision + "\t" + ndcg + "\t" + precision_tpk + "\t" + ndcg_tpk, file=log_file)
        log_file.flush()
        self.tic = time.time()
        self.nr_print += 1
        

    def cal_grad(self, input, labels, weights, trainR, tpk_idx, tpk_mask, pl_mask):
        # before estimation starts, need to zero out grad
        self.sess.run(self.zero_g)

        u,v = self.u, self.v
        p,q = self.p, self.q
        batch_grad = self.batch_grad
        loss = self.loss

        u_input = input.transform(input.feat)
        pl_input = input.transform(pl_mask)

        v_input = trainR.transform(trainR.feat)

        loss, _ = self.sess.run([loss, batch_grad], feed_dict={u: u_input, v: v_input,
            self.label_idx:labels, self.weights: weights, self.tpk_idx: tpk_idx,
            self.tpk_mask: tpk_mask, self.pl_mask: pl_input})

        self.sess.run(self.reg_grad)
        return loss


    def _update_model(self):
        update_model_ops = []

        x = inverse_vectorize(self.s, self.param)
        for i, p in enumerate(self.param):
            op = tf.assign(p, p + self.config.lr * x[i])
            update_model_ops.append(op)

        return tf.group(*update_model_ops)



    def train(self, trainL, trainR, validL, validR, saver):
        tf.summary.scalar('loss', self.f)
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter('./summary/train', self.sess.graph)

        print(self.config.args)
        if not self.config.screen_log_only:
            log_file = open(self.config.log_file, 'w')
            print(self.config.args, file=log_file)
        
        trainR.label = trainL.label.transpose()


        if self.config.lr == 0:
            raise ValueError('Learning rate  cannot be zero!')

        total_running_time = 0.0
        self.config.elapsed_time = 0.0
        u, v, p, q = self.network
        self.tic = time.time()
        for k in range(self.config.iter_max):

            nb_i = math.ceil(self.config.m / self.config.bsize_i)
            num_segment = nb_i
            for b in range(num_segment):

                idx_i = np.random.choice(np.arange(0, self.config.m),
                        size=self.config.bsize_i, replace=False)

                s = time.time()

                rows, cols = trainL.label.Label[idx_i].nonzero()
                labels = np.array([rows, cols]).transpose()
                weights = trainL.label.Weight[idx_i].data
                tpk_idx = trainL.label.tpk_idx[idx_i] if trainL.label.tpk_idx is not None else np.zeros((self.config.bsize_i, self.config.tpk_num))
                tpk_mask = trainL.label.tpk_mask[idx_i] if trainL.label.tpk_mask is not None else np.zeros((self.config.bsize_i, self.config.tpk_num))

                loss = self.cal_grad(trainL[idx_i], labels, weights, trainR, tpk_idx, tpk_mask, trainL.label.pl_mask[idx_i])
                reg = self.sess.run(self.cal_reg)
                self.f = loss + reg

                self.sess.run(self.update_adagrad)
                self.sess.run(self.gradient_direction)
                self.sess.run(self.update_model)

                if b % max(1, round(num_segment/10)) == 0:
                    self.print_info(k, b, num_segment, trainL, trainR, validL, validR, log_file)

        self.print_info(self.config.iter_max, 0, num_segment, trainL, trainR, validL, validR, log_file)
        if self.config.args.save_pq:
            save_PQ(self.sess, self.network, validL, validR, self.config, self.dtype_np)
            precision, ndcg, precision_tpk, ndcg_tpk = predict(self.sess, self.network, validL, validR, self.config, self.dtype_np)
            precision = "p@k"+ " " + " ".join([str(round(i*100,3)) for i in precision])
            ndcg = "ndcg@k" + " " + " ".join([str(round(i*100,3)) for i in ndcg])
            precision_tpk = "p@k_tpk"+ " " + " ".join([str(round(i*100,3)) for i in precision_tpk])
            ndcg_tpk = "ndcg@k_tpk" + " " + " ".join([str(round(i*100,3)) for i in ndcg_tpk])
            print("iter: %d "%k + precision + "\t" + ndcg + "\t" + precision_tpk + "\t" + ndcg_tpk)
        if not self.config.screen_log_only:
            log_file.close()
